#include <stdio.h>
#include <stdint.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "md5.h"

#define IFNAMSIZ	16

static unsigned int
lastchar(const unsigned char *str, unsigned int len)
{
	return str[len-1];
}

static unsigned int
kr_hash(const unsigned char *str, unsigned int len)
{
	unsigned int hash = 0;

        while (len--)
	    hash += *str++;

	return hash;
}

static unsigned int
sdbm(const unsigned char *name, unsigned int len)
{
        unsigned long hash = 0;

        while (len--)
            hash = *name++ + (hash << 6) + (hash << 16) - hash;

        return hash;
}

#define init_name_hash()                0

/* partial hash update function. Assume roughly 4 bits per character */
static inline unsigned long
partial_name_hash(unsigned long c, unsigned long prevhash)
{
        return (prevhash + (c << 4) + (c >> 4)) * 11;
}

/* Compute the hash for a name string. */
static unsigned int
full_name_hash(const unsigned char *name, unsigned int len)
{
        unsigned long hash = 0;
        while (len--)
                hash = partial_name_hash(*name++, hash);
        return hash;
}

static unsigned int
djb2(const unsigned char *str, unsigned int len)
{
	unsigned long hash = 5381;

        while (len--)
		hash = ((hash << 5) + hash) + *str++; /* hash * 33 + c */

        return hash;
}

static unsigned int
string_hash31(const unsigned char *name, unsigned int len)
{
        unsigned hash = 0;
        int i;

        for (i = 0; i < len; ++i)
                hash = 31 * hash + name[i];
        return hash;
}

static unsigned int
string_hash17(const unsigned char *name, unsigned int len)
{
        unsigned hash = 0;
        int i;

        for (i = 0; i < len; ++i)
                hash = 17 * hash + name[i];
        return hash;
}

static unsigned int 
string10(const unsigned char *name, unsigned int len)
{
        unsigned hash = 0;
        int i;

        for (i = 0; i < len; ++i)
                hash = 10 * hash + name[i] - '0';
        return hash;
}

/* will need to change to unaligned_access() */
static inline uint16_t get16bits(const unsigned char *data)
{
	return *(uint16_t *) data;
}

static inline unsigned int
SuperFastHash (const unsigned char * data, unsigned int len)
{
	uint32_t hash = len, tmp;
	int rem;

	rem = len & 3;
	len >>= 2;

	/* Main loop */
	for (;len > 0; len--) {
		hash  += get16bits (data);
		tmp    = (get16bits (data+2) << 11) ^ hash;
		hash   = (hash << 16) ^ tmp;
		data  += 2*sizeof (uint16_t);
		hash  += hash >> 11;
	}

	/* Handle end cases */
	switch (rem) {
        case 3: hash += get16bits (data);
                hash ^= hash << 16;
                hash ^= data[sizeof (uint16_t)] << 18;
                hash += hash >> 11;
                break;
        case 2: hash += get16bits (data);
                hash ^= hash << 11;
                hash += hash >> 17;
                break;
        case 1: hash += *data;
                hash ^= hash << 10;
                hash += hash >> 1;
	}

	/* Force "avalanching" of final 127 bits */
	hash ^= hash << 3;
	hash += hash >> 5;
	hash ^= hash << 4;
	hash += hash >> 17;
	hash ^= hash << 25;
	hash += hash >> 6;

	return hash;
}


/* NOTE: Arguments are modified. */
#define __jhash_mix(a, b, c) \
{ \
  a -= b; a -= c; a ^= (c>>13); \
  b -= c; b -= a; b ^= (a<<8); \
  c -= a; c -= b; c ^= (b>>13); \
  a -= b; a -= c; a ^= (c>>12);  \
  b -= c; b -= a; b ^= (a<<16); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>3);  \
  b -= c; b -= a; b ^= (a<<10); \
  c -= a; c -= b; c ^= (b>>15); \
}

/* The golden ration: an arbitrary value */
#define JHASH_GOLDEN_RATIO	0x9e3779b9

/* The most generic version, hashes an arbitrary sequence
 * of bytes.  No alignment or length assumptions are made about
 * the input key.
 */
static inline uint32_t jhash(const void *key, uint32_t length, uint32_t initval)
{
	uint32_t a, b, c, len;
	const uint8_t *k = key;

	len = length;
	a = b = JHASH_GOLDEN_RATIO;
	c = initval;

	while (len >= 12) {
		a += (k[0] +((uint32_t)k[1]<<8) +((uint32_t)k[2]<<16) +((uint32_t)k[3]<<24));
		b += (k[4] +((uint32_t)k[5]<<8) +((uint32_t)k[6]<<16) +((uint32_t)k[7]<<24));
		c += (k[8] +((uint32_t)k[9]<<8) +((uint32_t)k[10]<<16)+((uint32_t)k[11]<<24));

		__jhash_mix(a,b,c);

		k += 12;
		len -= 12;
	}

	c += length;
	switch (len) {
	case 11: c += ((uint32_t)k[10]<<24);
	case 10: c += ((uint32_t)k[9]<<16);
	case 9 : c += ((uint32_t)k[8]<<8);
	case 8 : b += ((uint32_t)k[7]<<24);
	case 7 : b += ((uint32_t)k[6]<<16);
	case 6 : b += ((uint32_t)k[5]<<8);
	case 5 : b += k[4];
	case 4 : a += ((uint32_t)k[3]<<24);
	case 3 : a += ((uint32_t)k[2]<<16);
	case 2 : a += ((uint32_t)k[1]<<8);
	case 1 : a += k[0];
	};

	__jhash_mix(a,b,c);

	return c;
}

static unsigned int
jhash_string(const unsigned char *name, unsigned int len)
{
	return jhash(name, len, 0);
}


static unsigned int
md5_string(const unsigned char *name, unsigned int len)
{
	MD5_CTX ctx;
	unsigned char digest[16];

	MD5Init(&ctx);
	MD5Update(&ctx, (unsigned char *) name, len);
	MD5Final(digest, &ctx);
	
	return *(unsigned int *)digest;

}

static unsigned int
fletcher(const unsigned char *data, unsigned int len)
{
	uint32_t sum1 = 0xffff, sum2 = 0xffff;
 
        while (len) {
                unsigned tlen = len > 360 ? 360 : len;
                len -= tlen;
                do {
                        sum1 += *data++;
                        sum2 += sum1;
                } while (--tlen);
                sum1 = (sum1 & 0xffff) + (sum1 >> 16);
                sum2 = (sum2 & 0xffff) + (sum2 >> 16);
        }
        /* Second reduction step to reduce sums to 16 bits */
        sum1 = (sum1 & 0xffff) + (sum1 >> 16);
        sum2 = (sum2 & 0xffff) + (sum2 >> 16);
        return sum2 << 16 | sum1;
}

#define MOD_ADLER 65521
 
static unsigned int adler32(const unsigned char *data, unsigned int len) 
{
    uint32_t a = 1, b = 0;
    unsigned int index;
 
    /* Loop over each byte of data, in order */
    for (index = 0; index < len; ++index) {
        a = (a + data[index]) % MOD_ADLER;
        b = (b + a) % MOD_ADLER;
    }
 
    return (b << 16) | a;
}


/*
 * There are multiple 16-bit CRC polynomials in common use, but this is
 * *the* standard CRC-32 polynomial, first popularized by Ethernet.
 * x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x^1+x^0
 */
#define CRCPOLY_LE 0xedb88320
#define CRCPOLY_BE 0x04c11db7

static unsigned int
crc(const unsigned char *p, unsigned int len)
{
	uint32_t crc = 0;
	unsigned int i;

	while (len--) {
		crc ^= *p++;
		for (i = 0; i < 8; i++)
			crc = (crc >> 1) ^ ((crc & 1) ? CRCPOLY_LE : 0);
	}
	return crc;
}

/* Peter Weinberger's */
static unsigned int pjw(const unsigned char *s, unsigned int len)
{
	uint32_t h = 0;

	while (len--) {
		uint32_t g;

    		h = (h<<4) + *s++;
    		g = h&0xF0000000;
		if (g) {
    			h ^= g>>24;
    			h ^= g;
    		}
    	}
    	return h;
}

/* UNIX ELF hash
 * Published hash algorithm used in the UNIX ELF format for object files
 */
static unsigned int elf(const unsigned char *name, unsigned int len)
{
    	uint32_t h = 0, g;
	
	while (len--) {
    		h = ( h << 4 ) + *name++;
    		g = h & 0xF0000000;
		if (g)
			h ^= g >> 24;
    		h &= ~g;
    	}

    	return h;
}

static unsigned int xor(const unsigned char *key, unsigned int len)
{
    	uint32_t h = 0;
	unsigned int i;
	
	for (i = 0; i < len; i++)
		h  = h ^ key[i];

    	return h;
}

static const uint32_t FNV_32_PRIME = 16777619u;
static const uint32_t FNV1_32_INIT = 2166136261u;

static unsigned int fnv32(const unsigned char *key, unsigned int len)
{
	uint32_t hval = FNV1_32_INIT;
	unsigned int i;

	for (i = 0; i < len; i++) {
		hval ^= key[i];
		hval *= FNV_32_PRIME;
	}

	return hval;
}

static const uint64_t FNV_64_PRIME = 1099511628211ull;
static const uint64_t FNV1_64_INIT = 14695981039346656037ull;

static unsigned int fnv64(const unsigned char *key, unsigned int len)
{
	uint64_t hval = FNV1_64_INIT;
	unsigned int i;

	for (i = 0; i < len; i++) {
		hval ^= key[i];
		hval *= FNV_64_PRIME;
	}

	return hval;
}

static unsigned testsize;
static unsigned hashbits;
static char *fmt = "eth%u";

static inline void measure(const char *name,  
			   unsigned int (*hash)(const unsigned char *, unsigned int))
{
	unsigned int i, h;
	struct timeval t0, t1;
	unsigned hashmsk = (1 << hashbits) - 1;
	unsigned long hashsz = hashmsk + 1;
	unsigned int dist[hashsz];
	unsigned long long probes = 0;
	long us;
	unsigned long m = 0;
	const double avg = testsize / hashsz;
	double ideal = hashsz * (avg * (1 + avg)) / 2;
	double std = 0;

	for (i = 0; i < hashsz; i++)
		dist[i] = 0;

	/* add loopback */
	h = hash((unsigned char *)"lo", 2);
	++dist[h & hashmsk];

	gettimeofday(&t0, NULL);
	for (i = 0; i < testsize; i += 10) {
		char name[IFNAMSIZ];
		int len = snprintf(name, IFNAMSIZ, fmt, i);
		unsigned int j = 0;

		do {
			h = hash((unsigned char *)name, len);
			++dist[h & hashmsk];
			name[len-1] = '0' + ++j;
		} while (j < 10);
	}

	gettimeofday(&t1, NULL);
	us = (t1.tv_sec - t0.tv_sec) * 1000000;
	us += (t1.tv_usec - t0.tv_usec);

	for (i = 0; i < hashsz; i++) {
		unsigned int n = dist[i];
		double delta = (n - avg);

		if (n > m) m = n;	/* longest chain */
		
		std += delta * delta;	/* compute standard deviation */
		
		/* number of probes used when accessing 
		   is same as sum of arithmetic series */
		probes += ((uint64_t) n * (1 + n)) / 2;
	}


	printf("%-20s %lu.%06lu", name, us / 1000000, us % 1000000);
	printf(" %10.2f %9lu %6.2f\n",
	       (double) probes / ideal, m, sqrt(std / testsize));

}
#define MEASURE(func)	measure(#func, &func)

static void usage(void)
{
	fprintf(stderr, "Usage:hashtest names hashbits\n");
	exit(1);
}

int main(int argc, char **argv)
{
	if (argc != 3)
		usage();

	testsize = strtoul(argv[1], NULL, 0);
	if (testsize == 0)
		usage();

	
	hashbits = strtoul(argv[2], NULL, 0);
	if (hashbits == 0) 
		usage();

	printf("Algorithm             Time       Ratio       Max   StdDev\n");
	MEASURE(lastchar);
	MEASURE(kr_hash);
	MEASURE(string_hash31);
	MEASURE(SuperFastHash);
	MEASURE(djb2);
	MEASURE(string_hash17);
	MEASURE(string10);
	MEASURE(full_name_hash);
	MEASURE(jhash_string);
	MEASURE(sdbm);
	MEASURE(md5_string);
	MEASURE(fletcher);
	MEASURE(adler32);
	MEASURE(crc);
	MEASURE(elf);
	MEASURE(pjw);
	MEASURE(xor);
	MEASURE(fnv32);
	MEASURE(fnv64);

        return 0;
}
